#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Libraries
import re
# Modules
from . import orchestrator

question_type = [
    (r"What is ([^\.\n]*) ", lambda matched_str, **kwargs: matched_what(matched_str, **kwargs)),

    (r"Is (.*) allocated ", lambda matched_str, **kwargs: matched_allocated(matched_str, **kwargs)),
]


def lookup(strings, questions, **kwargs):
    """Method to match regex"""
    answer_list = []
    for pattern, values in questions:
        for i in strings:
            result = re.findall(pattern, i, re.MULTILINE)
            if result:
                answer = values(result, **kwargs)
                if answer:
                    answer_list.append(answer)
    return answer_list


def find_question(string, **kwargs):
    """Entry point from jarvis.py"""
    out = lookup(string, question_type, **kwargs)
    return out


def get_objects_name_lists(**kwargs):
    """Returns lists of objects with their names depending on kwargs"""
    xml_function_name_list, xml_data_name_list, xml_state_name_list = [], [], []
    xml_fun_elem_name_list, xml_transition_name_list, xml_fun_inter_name_list = [], [], []

    # Create object names/aliases lists
    if kwargs.get('xml_function_list', False):
        xml_function_name_list = orchestrator.get_object_name(kwargs['xml_function_list'])
    if kwargs.get('xml_data_list', False):
        xml_data_name_list = orchestrator.get_object_name(kwargs['xml_data_list'])
    if kwargs.get('xml_state_list', False):
        xml_state_name_list = orchestrator.get_object_name(kwargs['xml_state_list'])
    if kwargs.get('xml_fun_elem_list', False):
        xml_fun_elem_name_list = orchestrator.get_object_name(kwargs['xml_fun_elem_list'])
    if kwargs.get('xml_transition_list', False):
        xml_transition_name_list = orchestrator.get_object_name(kwargs['xml_transition_list'])
    if kwargs.get('xml_fun_inter_list', False):
        xml_fun_inter_name_list = orchestrator.get_object_name(kwargs['xml_fun_inter_list'])

    whole_objects_name_list = [xml_function_name_list,
                               xml_data_name_list,
                               xml_state_name_list,
                               xml_fun_elem_name_list,
                               xml_transition_name_list,
                               xml_fun_inter_name_list]
    return whole_objects_name_list


def check_get_object(object_str, **kwargs):
    """
    Returns the desired object from object's string
    Args:
        object_str ([object_string]): list of object's name from cell
        **kwargs: xml lists

    Returns:
        wanted_object : Function/State/Data/Fun_Elem/Transition/Fun_Inter
    """
    whole_objects_name_list = get_objects_name_lists(**kwargs)

    if not [object_str in s for s in whole_objects_name_list]:
        print(f"{object_str} does not exist")
    else:
        result_function = [object_str in whole_objects_name_list[0]]
        result_data = [object_str in whole_objects_name_list[1]]
        resul_state = [object_str in whole_objects_name_list[2]]
        result_fun_elem = [object_str in whole_objects_name_list[3]]
        result_transition = [object_str in whole_objects_name_list[4]]
        result_fun_inter = [object_str in whole_objects_name_list[5]]
        result = [*result_function, *result_data, *resul_state,  *result_fun_elem,
                  *result_transition, *result_fun_inter]
        wanted_object = match_object(object_str, result, **kwargs)
        return wanted_object


def matched_what(question_str, **kwargs):
    """Get the 'what' declaration """
    object_str = question_str[0].strip()
    wanted_object = check_get_object(object_str, **kwargs)

    if wanted_object:
        object_info = get_object_info(wanted_object, **kwargs)
        if object_info:
            return object_info


# TODO: Use check_get_object and this method within orchestrator.py
def match_object(object_str, result, **kwargs):
    """Returns wanted_object from object_str and result matched from name lists"""
    if result[0]:
        for function in kwargs['xml_function_list']:
            if object_str in (function.name, function.alias):
                return function
    elif result[1]:
        for data in kwargs['xml_data_list']:
            if object_str == data.name:
                return data
    elif result[2]:
        for state in kwargs['xml_state_list']:
            if object_str in (state.name, state.alias):
                return state
    elif result[3]:
        for fun_elem in kwargs['xml_fun_elem_list']:
            if object_str in (fun_elem.name, fun_elem.alias):
                return fun_elem
    elif result[4]:
        for transition in kwargs['xml_transition_list']:
            if object_str in (transition.name, transition.alias):
                return transition
    elif result[5]:
        for fun_inter in kwargs['xml_fun_inter_list']:
            if object_str in (fun_inter.name, fun_inter.alias):
                return fun_inter


def get_object_info(wanted_object, **kwargs):
    """From wanted_object and all lists, returns a dict with info"""
    object_info = {'Name': str(wanted_object.name),
                   'Object Class': str(type(wanted_object).__name__),
                   'Type': str(wanted_object.type), 'Id': str(wanted_object.id)}

    try:
        # Only Data() does not have these attributes and Transition() does not have parent
        if any(wanted_object.alias):
            object_info['Alias'] = str(wanted_object.alias)
        if wanted_object.parent is not None:
            object_info['Parent'] = str(wanted_object.parent.name)
    except AttributeError:
        # To avoid error when there is no such attribute for the object
        pass
    if object_info['Object Class'] == 'Function':
        get_function_info(wanted_object, object_info, **kwargs)
    elif object_info['Object Class'] == 'FunctionalElement':
        get_fun_elem_info(wanted_object, object_info, **kwargs)
    elif object_info['Object Class'] == 'State':
        if any(wanted_object.child_list):
            object_info['Child List'] = get_child_name_list(wanted_object, kwargs['xml_state_list'])
    elif object_info['Object Class'] == 'Transition':
        get_transition_info(wanted_object, kwargs['xml_state_list'], object_info)
    elif object_info['Object Class'] == 'Data':
        get_data_info(wanted_object, object_info, **kwargs)

    if object_info:
        return object_info

    return


def get_function_info(wanted_object, object_info, **kwargs):
    """Get Function info"""
    if any(wanted_object.child_list):
        object_info['Child List'] = get_child_name_list(wanted_object, kwargs['xml_function_list'])
    if wanted_object.input_role is not None:
        object_info['Input Role'] = str(wanted_object.input_role)
    if wanted_object.operand is not None:
        object_info['Operand'] = str(wanted_object.operand)
    if any(wanted_object.port_list):
        object_info['Port List'] = str(wanted_object.port_list)

    object_info['Consumption List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_consumer_function_list'])
    object_info['Production List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_producer_function_list'])
    return object_info


def get_consumes_produces_info(wanted_object, relationship_list):
    """Get consumes/produces info"""
    object_relationship = set()
    for elem in relationship_list:
        if elem[1].id == wanted_object.id:
            object_relationship.add(elem[0])
        if elem[0] == wanted_object.name:
            object_relationship.add(elem[1].name)
    if object_relationship:
        return object_relationship


def get_child_name_list(parent_object, object_list):
    """Get child's name list"""
    child_list = set()
    for child in object_list:
        if child in parent_object.child_list:
            child_list.add((child.name, "Child"))
    return child_list


def get_allocated_object_name_list(wanted_object, object_list):
    """From Functional Element object and list (State/Function) get the allocated objects"""
    allocation_list = set()
    for allocated_object in object_list:
        if any(s == allocated_object.id for s in wanted_object.allocated_state_list):
            allocation_list.add(allocated_object.name)
        if any(s == allocated_object.id for s in wanted_object.allocated_function_list):
            allocation_list.add(allocated_object.name)
    return allocation_list


def get_transition_info(wanted_object, state_list, object_info):
    """Get transition's info"""
    for state in state_list:
        if state.id == wanted_object.source:
            object_info['Source'] = str(state.name)
        if state.id == wanted_object.destination:
            object_info['Destination'] = str(state.name)
    if wanted_object.condition_list:
        object_info['Condition List'] = wanted_object.condition_list
    return object_info


def get_fun_elem_info(wanted_object, object_info, **kwargs):
    """Get functional element info, i.e. child_list, allocated Function/State"""
    if any(wanted_object.child_list):
        object_info['Child List'] = get_child_name_list(wanted_object, kwargs['xml_fun_elem_list'])
    if any(wanted_object.allocated_state_list):
        object_info['Allocated State List'] = \
            get_allocated_object_name_list(wanted_object, kwargs['xml_state_list'])
    if any(wanted_object.allocated_function_list):
        object_info['Allocated Function List'] = \
            get_allocated_object_name_list(wanted_object, kwargs['xml_function_list'])
    return object_info


def get_data_info(wanted_object, object_info, **kwargs):
    """Get what consumes/produces for a specific Data object"""
    pred_list = set()
    object_info['Consumer List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_consumer_function_list'])
    object_info['Producer List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_producer_function_list'])
    if any(wanted_object.predecessor_list):
        for pred in wanted_object.predecessor_list:
            pred_list.add(pred.name)
    object_info['Predecessor List'] = pred_list
    return object_info


def matched_allocated(object_str, **kwargs):
    """From object_str, get object then check if allocated and returns allocation's list"""
    object_info = ""
    object_str = object_str[0]
    xml_function_name_list = orchestrator.get_object_name(kwargs['xml_function_list'])
    xml_state_name_list = orchestrator.get_object_name(kwargs['xml_state_list'])
    whole_objects_name_list = [*xml_function_name_list, *xml_state_name_list]
    if not any(s == object_str for s in whole_objects_name_list):
        print(f"{object_str} is not a function nor a state")
    else:
        result_function = any(s == object_str for s in xml_function_name_list)
        resul_state = any(s == object_str for s in xml_state_name_list)
        result = [result_function, False, resul_state,  False, False]
        wanted_object = match_object(object_str, result, **kwargs)
        if wanted_object:
            allocation_list = get_allocation_object(wanted_object, kwargs['xml_fun_elem_list'])
            if allocation_list:
                object_info += f'"{wanted_object.name}" is allocated to ' \
                               + ", ".join([elem.name for elem in allocation_list])
                return object_info

    return


def get_allocation_object(wanted_object, object_list):
    """Get current allocation for an object Fun_elem with State/Function OR Fun_inter with data"""
    allocation_list = set()
    object_type = orchestrator.get_object_type(wanted_object)

    if object_type == 'function':
        for fun_elem in object_list:
            if any(s == wanted_object.id for s in fun_elem.allocated_function_list):
                allocation_list.add(fun_elem)
    elif object_type == 'state':
        for fun_elem in object_list:
            if any(s == wanted_object.id for s in fun_elem.allocated_state_list):
                allocation_list.add(fun_elem)
    elif object_type == 'data':
        for fun_inter in object_list:
            if any(s == wanted_object.id for s in fun_inter.allocated_data_list):
                allocation_list.add(fun_inter)
    elif object_type == 'Functional interface':
        for fun_elem in object_list:
            if any(s == wanted_object.id for s in fun_elem.exposed_interface_list):
                allocation_list.add(fun_elem)
    elif object_type == 'Functional element':
        for function in object_list:
            if any(s == function.id for s in wanted_object.allocated_function_list):
                allocation_list.add(function)
    if allocation_list:
        return allocation_list

    return


def switch_objects_lists(type_list_str, wanted_object, object_type, **kwargs):
    """Switch depending on list's type and object's type """
    if object_type in ("state", "function", "Functional element"):
        switch_type_list = {
            "input": switch_in,
            "output": switch_out,
            "child": switch_child,
            "function": switch_state_function,
            "transition": switch_state_transition,
            "interface": switch_fun_elem_interface,
        }
        if object_type == "state" and type_list_str in ("input", "output"):
            return case_no_list(wanted_object, object_type, **kwargs)
        elif object_type != "state" and type_list_str in ("function", "transition"):
            return case_no_list(wanted_object, object_type, **kwargs)
        elif object_type != "Functional element" and type_list_str == "interface":
            return case_no_list(wanted_object, object_type, **kwargs)
        else:
            type_list = switch_type_list.get(type_list_str, case_no_list)
            return type_list(wanted_object, object_type, **kwargs)

    elif object_type == "Functional interface" and type_list_str == "data":
        return switch_data(wanted_object, object_type, **kwargs)

    else:
        return case_no_list(wanted_object, object_type, **kwargs)


def switch_in(wanted_object, object_type, **kwargs):
    """Case 'list input Function/Functional ELement' """
    input_list = get_input(wanted_object, **kwargs)
    if input_list:
        list_name = f"Input list for {wanted_object.name}:"
        input_list.insert(0, list_name)
        return input_list


def switch_out(wanted_object, object_type, **kwargs):
    """Case 'list output Function/Functional ELement' """
    output_list = get_output(wanted_object, **kwargs)
    if output_list:
        list_name = f"Output list for {wanted_object.name}:"
        output_list.insert(0, list_name)
        return output_list


def switch_child(wanted_object, object_type, **kwargs):
    """Case 'list child Function/State/Functional ELement' """
    child_list = None
    if object_type == "function":
        child_list = list(get_child_name_list(wanted_object,
                                              kwargs['xml_function_list']))
    elif object_type == "state":
        child_list = list(get_child_name_list(wanted_object,
                                              kwargs['xml_state_list']))
    elif object_type == "Functional element":
        child_list = list(get_child_name_list(wanted_object,
                                              kwargs['xml_fun_elem_list']))
        for allocated_fun in wanted_object.allocated_function_list:
            for fun in kwargs['xml_function_list']:
                if fun.id == allocated_fun:
                    child_list.append((fun.name, "Function allocation"))
        for allocated_state in wanted_object.allocated_state_list:
            for state in kwargs['xml_state_list']:
                if state.id == allocated_state:
                    child_list.append((state.name, "State allocation"))
    if child_list:
        list_name = f"Child list for {wanted_object.name}:"
        child_list = list(tuple(sorted(child_list)))
        child_list.insert(0, list_name)

        return child_list


def switch_state_function(wanted_object, object_type, **kwargs):
    """Case 'list function State' """

    function_list = []
    for allocated_fun in wanted_object.allocated_function_list:
        for fun in kwargs['xml_function_list']:
            if fun.id == allocated_fun:
                function_list.append((fun.name, "Function allocation"))

    if function_list:
        list_name = f"Function list for {wanted_object.name}:"
        function_list = list(tuple(sorted(function_list)))
        function_list.insert(0, list_name)

        return function_list


def switch_state_transition(wanted_object, object_type, **kwargs):
    transition_list = []
    for transition in kwargs['xml_transition_list']:
        if wanted_object.id == transition.source:
            for state in kwargs['xml_state_list']:
                if transition.destination == state.id:
                    transition_list.append({
                        'Transition name': transition.name,
                        'Source state': wanted_object.name,
                        'Destination state': state.name,
                        'Condition(s)': transition.condition_list
                    })
        elif wanted_object.id == transition.destination:
            for state in kwargs['xml_state_list']:
                if transition.source == state.id:
                    transition_list.append({
                        'Transition name': transition.name,
                        'Source state': state.name,
                        'Destination state': wanted_object.name,
                        'Condition(s)': transition.condition_list
                    })

    if transition_list:
        list_name = f"Transition list for {wanted_object.name}:"
        transition_list.insert(0, list_name)

        return transition_list


def switch_fun_elem_interface(wanted_object, object_type, **kwargs):
    """Case for 'list interface Functional element'"""
    fun_inter_list = get_objects_from_id_list([i for i in wanted_object.exposed_interface_list],
                                              kwargs['xml_fun_inter_list'])
    main_fun_elem_child_list, _ = orchestrator.get_children(wanted_object)
    if not fun_inter_list:
        return f"Not any exposed interface for {wanted_object.name}"
    else:
        parent_fun_elem = set()
        for interface in fun_inter_list:
            for fun_elem in kwargs['xml_fun_elem_list']:
                if fun_elem not in main_fun_elem_child_list and \
                        interface.id in fun_elem.exposed_interface_list:
                    parent_fun_elem.add((interface, fun_elem))

        interface_list = set()
        for k in parent_fun_elem:
            child_list, _ = orchestrator.get_children(k[1])
            child_list.remove(k[1])
            if child_list:
                check = True
                for p in parent_fun_elem:
                    if p[0] == k[0] and any(a == p[1] for a in child_list):
                        check = False
                        break
                if check:
                    interface_list.add((k[0].name, k[1].name))
            else:
                interface_list.add((k[0].name, k[1].name))
        if interface_list:
            list_name = f"Interface list for {wanted_object.name}:"
            interface_list = list(tuple(sorted(interface_list)))
            interface_list.insert(0, list_name)
            return interface_list


def switch_data(wanted_object, object_type, **kwargs):
    """Case for 'list data Functional Interface' """
    data_list = []
    fun_elem_exposing = get_allocation_object(wanted_object, kwargs['xml_fun_elem_list'])
    last_fun_elem_exposing = [check_latest(j, fun_elem_exposing) for j in fun_elem_exposing]
    for allocated_id in wanted_object.allocated_data_list:
        for data in kwargs['xml_data_list']:
            if allocated_id == data.id:
                data_list.append(get_latest_obj_interface(data, last_fun_elem_exposing, **kwargs))

    if data_list:
        list_name = f"Data list for {wanted_object.name}:"
        data_list.insert(0, list_name)
        return data_list


def case_no_list(wanted_object, object_type, **kwargs):
    """Case when there is incompatible list's type with object's type """
    return f"No list available for object '{wanted_object.name}' " \
           f"of type '{object_type.capitalize()}', possible lists are:\n" \
           f"- List child [Function/State/Functional element]\n" \
           f"- List input/output [Function/Functional element]\n" \
           f"- List function/transition [State]\n" \
           f"- List interface [Functional element]\n" \
           f"- List data [Functional interface]"


def get_latest_obj_interface(data, last_fun_elem_exposing, **kwargs):
    """For a data, find last producer and consumer and if they are allocated to last fun_elem
    exposing the functional interface asked"""
    data_dict = {'Data': data.name,
                 'Last consumer Function(s)': [],
                 'Last consumer Functional element(s)': [],
                 'Last producer Function(s)': [],
                 'Last producer Functional element(s)': []}
    for prod in kwargs['xml_producer_function_list']:
        if prod[0] == data.name and \
                check_latest(prod[1], kwargs['xml_function_list']) == prod[1].name:
            for cons in kwargs['xml_consumer_function_list']:
                cons_last_fun_elem = None
                prod_last_fun_elem = None
                if cons[0] == prod[0] and \
                        check_latest(cons[1], kwargs['xml_function_list']) == cons[1].name:
                    cons_fun_elem_list = get_allocation_object(cons[1], kwargs['xml_fun_elem_list'])
                    if cons_fun_elem_list:
                        for fun_elem in cons_fun_elem_list:
                            if fun_elem.name in last_fun_elem_exposing:
                                cons_last_fun_elem = fun_elem.name
                    prod_fun_elem_list = get_allocation_object(prod[1], kwargs['xml_fun_elem_list'])
                    if prod_fun_elem_list:
                        for fun_elem in prod_fun_elem_list:
                            if fun_elem.name in last_fun_elem_exposing:
                                prod_last_fun_elem = fun_elem.name

                if not cons_last_fun_elem == prod_last_fun_elem:
                    if not any(c == prod[1].name for c in data_dict['Last producer Function(s)']):
                        data_dict['Last producer Function(s)'].append(prod[1].name)
                    if prod_last_fun_elem is None:
                        data_dict['Last producer Functional element(s)'].append(
                            f"'{prod[1].name}' is not allocated")
                    else:
                        data_dict['Last producer Functional element(s)'].append(prod_last_fun_elem)
                    if not any(c == cons[1].name for c in data_dict['Last consumer Function(s)']):
                        data_dict['Last consumer Function(s)'].append(cons[1].name)
                    if cons_last_fun_elem is None:
                        data_dict['Last consumer Functional element(s)'].append(
                            f"'{cons[1].name}' is not allocated")
                    else:
                        data_dict['Last consumer Functional element(s)'].append(cons_last_fun_elem)

    return data_dict


def check_latest(wanted_object, check_list):
    """Checks and returns latest object = last one decomposed"""
    if not wanted_object.child_list:
        return wanted_object.name
    else:
        check_child = [c for c in wanted_object.child_list]
        if not any(j in check_child for j in check_list):
            return wanted_object.name


def get_object_list(object_str, **kwargs):
    """
    Gets lists from object_str : i.e. Input/Output/Child/Data
    Args:
        object_str ([object_string]): list of object's name from cell
        **kwargs: all xml lists

    Returns:
        answer_list: [Input_list, Output_list, Child_list, Data_list]
    """
    answer_list = []
    for elem in object_str:
        wanted_object = check_get_object(elem[1], **kwargs)
        if wanted_object is None:
            print(f"Object '{elem[1]}' does not exist")
        else:
            object_type = orchestrator.get_object_type(wanted_object)
            wanted_list = switch_objects_lists(elem[0], wanted_object, object_type, **kwargs)
            if isinstance(wanted_list, list):
                answer_list.append(wanted_list)
            elif isinstance(wanted_list, str):
                print(wanted_list)
            else:
                print(f"Nothing to display for {elem[0]} list of '{wanted_object.name}'")

    return answer_list


# TODO: Merge get_output and get_input()
def get_input(wanted_object, unmerged=False, **kwargs):
    """
    Gets inputs for object (Function or Functional Element): i.e. what is consumed by wanted object.
    Args:
        wanted_object: current object
        unmerged (bool: default=false): used by functional elements
        **kwargs: all xml lists

    Returns:
        Input's list
    """
    object_type = orchestrator.get_object_type(wanted_object)
    if object_type == "Functional element":
        input_list = []
        allocated_fun_list = set()
        for fun in wanted_object.allocated_function_list:
            for xml_fun in kwargs['xml_function_list']:
                if fun == xml_fun.id:
                    allocated_fun_list.add(xml_fun)
        for func in allocated_fun_list:
            current_fun_list = get_input(func, True, **kwargs)
            for sub in current_fun_list:
                if sub and sub[1] not in [f.name for f in allocated_fun_list]:
                    input_list.append(sub)
        return merge_list_per_cons_prod(input_list)
    else:
        input_list = get_in_out_function(wanted_object,
                                         kwargs['xml_consumer_function_list'],
                                         kwargs['xml_producer_function_list'])
        if unmerged:
            return input_list
        else:
            return merge_list_per_cons_prod(input_list)


def get_output(wanted_object, unmerged=False, **kwargs):
    """
    Gets outputs for object (Function or Functional Element): i.e. what is produced by wanted
    object.
    Args:
        wanted_object: current object
        unmerged (bool: default=false): used by functional elements
        **kwargs: all xml lists

    Returns:
        Output's list
    """
    object_type = orchestrator.get_object_type(wanted_object)
    if object_type == "Functional element":
        output_list = []
        allocated_fun_list = set()
        for fun in wanted_object.allocated_function_list:
            for xml_fun in kwargs['xml_function_list']:
                if fun == xml_fun.id:
                    allocated_fun_list.add(xml_fun)
        for func in allocated_fun_list:
            current_fun_list = get_output(func, True, **kwargs)
            for sub in current_fun_list:
                if sub and sub[1] not in [f.name for f in allocated_fun_list]:
                    output_list.append(sub)
        return merge_list_per_cons_prod(output_list)
    else:
        output_list = get_in_out_function(wanted_object,
                                          kwargs['xml_producer_function_list'],
                                          kwargs['xml_consumer_function_list'])
        if unmerged:
            return output_list
        else:
            return merge_list_per_cons_prod(output_list)


def merge_list_per_cons_prod(input_list):
    """
    Sorts data's name in alphabetical order and merges list by producer or consumer
    Args:
        input_list ([Data_name, funcion_name]): List of consumer/producer with data per Function

    Returns:
        Sorted + merged list
    """
    input_list = sorted(input_list)
    output_list = []
    empty_dict = {}
    for data, obj_name in input_list:
        if data not in empty_dict:
            output_list.append([data, obj_name])
            empty_dict[data] = len(empty_dict)
        else:
            if obj_name:
                if obj_name not in output_list[empty_dict[data]][1]:
                    output_list[empty_dict[data]][1] += '\\n' + obj_name

    return output_list


def get_in_out_function(wanted_object, wanted_relationship, opposite_wanted_relationship):
    """
    Return the list of Input/Output for Function's objects.
    Args:
        wanted_object (Function): Object's wanted
        wanted_relationship ([Xml_list]): For Output (producer's list) and Input (consumer's list)
        opposite_wanted_relationship ([Xml_list]): slef-explained
    Returns:
        output_list ([Data, prod/cons])
    """
    output_list = []
    flow_list = get_consumes_produces_info(wanted_object, wanted_relationship)
    if not flow_list:
        return output_list
    for data in flow_list:
        if not any(data in s for s in opposite_wanted_relationship):
            output_list.append([data, None])
        else:
            for elem in opposite_wanted_relationship:
                check = True
                if elem[0] == data:
                    for child in elem[1].child_list:
                        if [elem[0], child] in opposite_wanted_relationship:
                            check = False
                    if check:
                        output_list.append([elem[0], elem[1].name])

    return output_list


def get_objects_from_id_list(id_list, object_list):
    """From a list of id (from same type) returns list of corresponding objects"""
    if not id_list:
        return
    else:
        output_list = set()
        for wanted_object in object_list:
            if wanted_object.id in id_list:
                output_list.add(wanted_object)

        if output_list:
            return output_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Libraries
import sys
from lxml import etree

# Modules
from . import xml_writer # noqa
sys.path.append("../datamodel")
import datamodel # noqa


def generate_xml(xml_file):
    xml = xml_writer.GenerateXML(xml_file)
    return xml


# TODO: Delete dict not useful -> TBC
def parse_xml(input_filename):
    # Parse the XML file
    tree = etree.parse(input_filename)
    # Get the XML tree
    root = tree.getroot()
    # looking for elements with tag "function" and create function objects and list
    function_list, function_parent_list = get_functions(root)
    # Create data(and set predecessors), consumers, producers lists
    data_list, producer_function_list, consumer_function_list = get_data(root, function_list)
    # Looking for elements with tag "state" and create state objects and list
    state_list, state_parent_dict = get_state(root)
    # Looking for elements with tag "transition"
    transition_list = get_transition(root)
    # Looking for functional elements
    functional_element_list, fun_elem_parent_dict = get_functional_element(root)
    # Looking for chains
    chain_list = get_chains(root)
    # Looking for attributes
    attribute_list = get_attributes(root)
    # Looking for functional interfaces
    functional_interface_list = get_functional_interface(root)

    all_lists = [function_list, consumer_function_list, producer_function_list,
                 function_parent_list, data_list, state_list, state_parent_dict, transition_list,
                 functional_element_list, fun_elem_parent_dict, chain_list, attribute_list,
                 functional_interface_list]
    return all_lists


def get_functions(root):
    function_list = set()
    parent_list = {}
    xml_function_list = root.iter('function')
    for xml_function in xml_function_list:
        # Instantiate functions and add them to a list
        function = datamodel.Function()
        function.set_id(xml_function.get('id'))
        function.set_name(xml_function.get('name'))
        function.set_alias(xml_function.get('alias'))
        function_type = datamodel.FunctionType.get_name(xml_function.get('type'))
        function.set_type(function_type)
        function.set_operand()

        function_list.add(function)
        # Looking for functions with "functionalPart" i.e childs and create a list
        xml_function_part_list = xml_function.iter('functionalPart')
        for xml_function_part in xml_function_part_list:
            parent_list[xml_function_part.get('id')] = function.id

    # Loop to set parent() and add_child() to functions
    for child_id in parent_list:
        for child_function in function_list:
            if child_function.id == child_id:
                for parent_function in function_list:
                    if parent_function.id == parent_list[child_id]:
                        child_function.set_parent(parent_function)
                        parent_function.add_child(child_function)
                        break
                break

    return function_list, parent_list


# TODO: Change consumer/producer list from [data_name, function] to [data, function] and
#  extend it to all scripts
def get_data(root, function_list):
    data_list = set()
    consumer_function_list = []
    producer_function_list = []

    xml_data_list = root.iter('data')
    for xml_data in xml_data_list:
        # Instantiate data and add it to a list
        data = datamodel.Data()
        data.set_id(xml_data.get('id'))
        data.set_name(xml_data.get('name'))
        data_type = datamodel.DataType.get_name(xml_data.get('type'))
        data.set_type(data_type)

        data_list.add(data)
        # looking for all elements with tag "consumer" and create a list [flow, consumer_function]
        xml_consumer_list = xml_data.iter('consumer')
        for xml_consumer in xml_consumer_list:
            for function in function_list:
                if xml_consumer.get('id') == function.id:
                    consumer_function_list.append([xml_data.get('name'), function])
                    if xml_consumer.get('role') != 'none':
                        function.set_input_role(xml_data.get('name'))
                    # Avoid to reset the input role once already set
                    elif function.input_role is None:
                        function.set_input_role(None)

        # looking for all elements with tag "producer" and create a list [flow, producer_function]
        xml_producer_list = xml_data.iter('producer')
        for xml_producer in xml_producer_list:
            for function in function_list:
                if xml_producer.get('id') == function.id:
                    producer_function_list.append([xml_data.get('name'), function])

    # Loop on the data_list once created to find the predecessor and add it to list
    for d in data_list:
        xml_data_list = root.iter('data')
        for xml_data in xml_data_list:
            if xml_data.get('id') == d.id:
                # looking for all elements with tag "predecessor" and create a
                # list [flow, producer_function]
                xml_predecessor_list = xml_data.iter('predecessor')
                for xml_predecessor in xml_predecessor_list:
                    for dodo in data_list:
                        if xml_predecessor.get('id') == dodo.id:
                            d.add_predecessor(dodo)

    return data_list, producer_function_list, consumer_function_list


def get_state(root):
    state_list = set()
    state_parent_dict = {}
    xml_state_list = root.iter('state')
    for xml_state in xml_state_list:
        # Instantiate states and add them to a list
        state = datamodel.State()
        state.set_id(xml_state.get('id'))
        state.set_name(xml_state.get('name'))
        state.set_alias(xml_state.get('alias'))
        state_type = datamodel.StateType.get_name(xml_state.get('type'))
        state.set_type(state_type)

        state_list.add(state)
        # Looking for states with "statePart" i.e child and create a list
        xml_state_part_list = xml_state.iter('statePart')
        for xml_state_part in xml_state_part_list:
            state_parent_dict[xml_state_part.get('id')] = state.id

        # Looking for allocated functions and add them to the state
        xml_allocated_function_list = xml_state.iter('allocatedFunction')
        for xml_allo_fun in xml_allocated_function_list:
            state.add_allocated_function(xml_allo_fun.get("id"))

    # Loop to set parent() and add_child() to states
    for child_id in state_parent_dict:
        for child_state in state_list:
            if child_state.id == child_id:
                for parent_state in state_list:
                    if parent_state.id == state_parent_dict[child_id]:
                        child_state.set_parent(parent_state)
                        parent_state.add_child(child_state)
                        break
                break

    return state_list, state_parent_dict


def get_transition(root):
    transition_list = set()
    xml_transition_list = root.iter('transition')
    for xml_transition in xml_transition_list:
        # Instantiate transitions and add them to a list
        transition = datamodel.Transition()
        transition.set_id(xml_transition.get('id'))
        transition.set_name(xml_transition.get('name'))
        transition.set_alias(xml_transition.get('alias'))
        transition_type = datamodel.TransitionType.get_name(xml_transition.get('type'))
        transition.set_type(transition_type)
        transition.set_source(xml_transition.get('source'))
        transition.set_destination(xml_transition.get('destination'))

        # Looking for conditions and add them to the transition
        xml_transition_condition_list = xml_transition.iter('condition')
        for xml_condition in xml_transition_condition_list:
            transition.add_condition(xml_condition.get("text"))

        transition_list.add(transition)

    return transition_list


def get_functional_element(root):
    functional_element_list = set()
    fun_elem_parent_dict = {}
    xml_functional_element_list = root.iter('functionalElement')
    for xml_func_elem in xml_functional_element_list:
        # Instantiate functional element and add them to a list
        fun_elem = datamodel.FunctionalElement()
        fun_elem.set_id(xml_func_elem.get('id'))
        fun_elem.set_name(xml_func_elem.get('name'))
        fun_elem.set_alias(xml_func_elem.get('alias'))
        fun_elem_type = datamodel.FunctionalElementType.get_name(xml_func_elem.get('type'))
        fun_elem.set_type(fun_elem_type)

        functional_element_list.add(fun_elem)
        # Looking for states with "functionalElementPart" i.e child and create a list
        xml_functional_part_list = xml_func_elem.iter('functionalElementPart')
        for xml_state_part in xml_functional_part_list:
            fun_elem_parent_dict[xml_state_part.get('id')] = fun_elem.id

        # Looking for allocated states and add them to the functional element
        xml_allocated_state_list = xml_func_elem.iter('allocatedState')
        for xml_allo_state in xml_allocated_state_list:
            fun_elem.add_allocated_state(xml_allo_state.get("id"))

        # Looking for allocated functions and add them to the functional element
        xml_allocated_function_list = xml_func_elem.iter('allocatedFunction')
        for xml_allo_fun in xml_allocated_function_list:
            fun_elem.add_allocated_function(xml_allo_fun.get("id"))

        # Looking for exposed interface and add them to the functional element
        xml_exposed_interface_list = xml_func_elem.iter('exposedInterface')
        for xml_exp_inter in xml_exposed_interface_list:
            fun_elem.add_exposed_interface(xml_exp_inter.get("id"))

    # Loop to set parent() and add_child() to fun elem
    for child_id in fun_elem_parent_dict:
        for child_state in functional_element_list:
            if child_state.id == child_id:
                for parent_state in functional_element_list:
                    if parent_state.id == fun_elem_parent_dict[child_id]:
                        child_state.set_parent(parent_state)
                        parent_state.add_child(child_state)
                        break
                break

    return functional_element_list, fun_elem_parent_dict


def get_chains(root):
    chain_list = set()
    xml_chain_list = root.iter('chain')
    for xml_chain in xml_chain_list:
        # Instantiate chain and add them to a list
        chain = datamodel.Chain()
        chain.set_id(xml_chain.get('id'))
        chain.set_name(xml_chain.get('name'))
        chain_type = datamodel.ChainType.get_name(xml_chain.get('type'))
        chain.set_type(chain_type)

        # Looking for allocated items and add them to the chain
        xml_allocated_item_list = xml_chain.iter('allocatedItem')
        for xml_allo_item in xml_allocated_item_list:
            chain.add_allocated_item(xml_allo_item.get("id"))

        chain_list.add(chain)

    return chain_list


def get_attributes(root):
    attribute_list = set()
    xml_attribute_list = root.iter('attribute')
    for xml_attribute in xml_attribute_list:
        # Instantiate Attribute and add them to a list
        attribute = datamodel.Attribute()
        attribute.set_id(xml_attribute.get('id'))
        attribute.set_name(xml_attribute.get('name'))
        attribute.set_alias(xml_attribute.get('alias'))
        # No AttritubeType defined in datamodel yet
        # attribute_type = datamodel.AttributeType.get_name(xml_attribute.get('type'))
        attribute.set_type(xml_attribute.get('type'))

        # Looking for described items and add them to the attribute
        xml_described_item_list = xml_attribute.iter('describedItem')
        for xml_described_item in xml_described_item_list:
            attribute.add_described_item((xml_described_item.get("id"),
                                          xml_described_item.get("value")))

        attribute_list.add(attribute)

    return attribute_list


def get_functional_interface(root):
    functional_interface_list = set()
    xml_fun_inter_list = root.iter('functionalInterface')
    for xml_fun_inter in xml_fun_inter_list:
        # Instantiate fun_inter and add them to a list
        fun_inter = datamodel.FunctionalInterface()
        fun_inter.set_id(xml_fun_inter.get('id'))
        fun_inter.set_name(xml_fun_inter.get('name'))
        fun_inter.set_type(xml_fun_inter.get('type'))
        fun_inter.set_alias(xml_fun_inter.get('alias'))

        # Looking for allocated data and add them to the fun inter
        xml_allocated_data_list = xml_fun_inter.iter('allocatedData')
        for xml_allo_data in xml_allocated_data_list:
            fun_inter.add_allocated_data(xml_allo_data.get("id"))

        functional_interface_list.add(fun_inter)

    return functional_interface_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Libraries
from lxml import etree


# Class to generate XML
class GenerateXML:
    def __init__(self, xml_file):
        # Initialize XML structure/tags
        self.root = etree.Element("funcArch")
        self.tags = ['functionList', 'dataList', 'stateList', 'transitionList',
                     'functionalElementList', 'chainList', 'attributeList',
                     'functionalInterfaceList']
        for tag in self.tags:
            etree.SubElement(self.root, tag)
        self.tree = etree.ElementTree(self.root)

        if len(xml_file) > 0:
            self.file = xml_file
        else:
            self.file = "Output.xml"

    # Method to write functions from function's list
    def write_function(self, function_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('functionList') is None:
                etree.SubElement(root, 'functionList')
            # Loop on each flow/data
            for function_list_tag in root.findall("./functionList"):
                # Loop on function's list
                for function in function_list:
                    function_tag = etree.SubElement(function_list_tag, "function",
                                                    {'id': function.id,
                                                     'name': function.name,
                                                     'type': str(function.type),
                                                     'alias': function.alias})

                    _functional_part_list_tag = etree.SubElement(function_tag, "functionalPartList")

        self.write()

    # Method to write function's type by list [function, type]
    def write_function_type(self, function_type_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on [function, type] list
            for function in function_type_list:
                for function_tag in root.findall(".//function[@id='" + function[0].id + "']"):
                    function_tag.set('type', function[1])
        self.write()

    # Method to write function's alias by list [function, alias]
    def write_function_alias(self, function_alias_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on [function, type] list
            for function in function_alias_list:
                for function_tag in root.findall(".//function[@id='" + function[0].id + "']"):
                    function_tag.set('alias', function[1])
        self.write()

    # Method to write child by list [parent, child]
    def write_function_child(self, child_function_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each flow/data
            for function in root.findall(".//function"):
                for parent, child in child_function_list:
                    if function.get('id') == parent.id:
                        tag = function.find('functionalPartList')
                        functional_part_tag = etree.SubElement(tag, "functionalPart",
                                                               {'id': child.id})
        self.write()

    # Method to add data flows
    def write_data(self, data_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('dataList') is None:
                etree.SubElement(root, 'dataList')
            for data_list_tag in root.findall('dataList'):
                for data in data_list:
                    existing_data_tag = data_list_tag.find('./dataList/data')
                    if existing_data_tag is not None:
                        tag = existing_data_tag
                    else:
                        tag = data_list_tag
                    data_tag = etree.SubElement(tag, "data", {'name': data.name,
                                                              'type': str(data.type),
                                                              'id': data.id})

                    _consumer_list_tag = etree.SubElement(data_tag, "consumerList")

                    _producer_list_tag = etree.SubElement(data_tag, "producerList")

                    _predecessor_list_tag = etree.SubElement(data_tag, "predecessorList")

        self.write()

    # Method to write data's type by list [data, type]
    def write_data_type(self, data_type_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on [function, type] list
            for data in data_type_list:
                for data_tag in root.findall(".//data[@name='" + data[0].name + "']"):
                    data_tag.set('type', data[1])
        self.write()

    # Method to write consumers by list [data_name, function]
    def write_consumer(self, consumer_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for i in consumer_list:
                for consumer_list_tag in root.findall("./dataList/data[@name='" + str(i[0])
                                                      + "']/consumerList"):
                    if not i[1].operand:
                        consumer_tag = etree.SubElement(consumer_list_tag, "consumer",
                                                        {'id': i[1].id, 'role': "none"})
                    else:
                        consumer_tag = etree.SubElement(consumer_list_tag, "consumer",
                                                        {'id': i[1].id, 'role': i[1].operand})

        self.write()

    # Method to write producers by list [data_name, function]
    def write_producer(self, producer_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for i in producer_list:
                for producer_list_tag in root.findall("./dataList/data[@name='" + str(i[0])
                                                      + "']/producerList"):

                    producer_tag = etree.SubElement(producer_list_tag, "producer", {'id': i[1].id})

        self.write()

    # Method to write predecessors by list [data, predecessor]
    def write_predecessor(self, predecessor_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for i in predecessor_list:
                for predecessor_list_tag in root.findall(
                        "./dataList/data[@name='" + str(i[0].name) + "']/predecessorList"):

                    _producer_tag = etree.SubElement(predecessor_list_tag, "predecessor",
                                                     {'id': i[1].id})

        self.write()

    # Method to remove function by list [function]
    def delete_function(self, delete_function_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for function in delete_function_list:
                for function_tag in root.findall(".//function[@id='" + function.id + "']"):
                    function_tag.getparent().remove(function_tag)
        self.write()

    # Method to remove data by list [data]
    def delete_data(self, delete_data_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for data in delete_data_list:
                for data_tag in root.findall(".//data[@name='" + data.name + "']"):
                    data_tag.getparent().remove(data_tag)
        self.write()

    # Method to delete the parents (consumer or producer) when flow is within a component
    def delete_single_consumer_producer(self, data, function, value):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Find specific consumer/producer to delete
            for tag in root.findall(
                    ".//data[@name='" + data + "']/" + value + "List/" + value + "[@id='"
                    + function.id + "']"):
                tag.getparent().remove(tag)
        self.write()

    # Method to write within XML file
    def write(self):
        with open(self.file, "wb") as file:
            self.tree.write(file, encoding='utf-8', xml_declaration=True, pretty_print=True)

    # Method to write states from state's list
    def write_state(self, state_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('stateList') is None:
                etree.SubElement(root, 'stateList')
            for state_list_tag in root.findall("./stateList"):
                for state in state_list:
                    state_tag = etree.SubElement(state_list_tag, "state",
                                                    {'id': state.id, 'name': state.name,
                                                     'type': str(state.type), 'alias': state.alias})

                    _state_part_list_tag = etree.SubElement(state_tag, "statePartList")
                    _allocated_function_list_tag = etree.SubElement(state_tag,
                                                                    "allocatedFunctionList")
        self.write()

    # Method to write state's type by list [state, type]
    def write_state_type(self, state_type_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for state in state_type_list:
                for state_tag in root.findall(".//state[@id='" + state[0].id + "']"):
                    state_tag.set('type', state[1])
        self.write()

    # Method to write state's alias by list [function, alias]
    def write_state_alias(self, state_alias_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for state in state_alias_list:
                for state_tag in root.findall(".//state[@id='" + state[0].id + "']"):
                    state_tag.set('alias', state[1])
        self.write()

    # Method to write child's state by list [parent, child]
    def write_state_child(self, child_state_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each state
            for state in root.findall(".//state"):
                for parent, child in child_state_list:
                    if state.get('id') == parent.id:
                        tag = state.find('statePartList')
                        state_part_tag = etree.SubElement(tag, "statePart",
                                                               {'id': child.id})
        self.write()

    # Method to write allocated function by list [state, allocated_function]
    def write_allocated_function_to_state(self, state_function_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each flow/data
            for xml_state in root.findall(".//state"):
                for fun_elem, function in state_function_list:
                    if xml_state.get('id') == fun_elem.id:
                        tag = xml_state.find('allocatedFunctionList')
                        allocated_state_tag = etree.SubElement(tag, "allocatedFunction",
                                                               {'id': function.id})
        self.write()

    # Method to delete state by list [state]
    def delete_state(self, delete_state_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for state in delete_state_list:
                for state_tag in root.findall(".//state[@id='" + state.id + "']"):
                    state_tag.getparent().remove(state_tag)
        self.write()

    # Method to write transition from transition's list
    def write_transition(self, transition_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('transitionList') is None:
                etree.SubElement(root, 'transitionList')
            for transition_list_tag in root.findall("./transitionList"):
                for transition in transition_list:
                    transition_tag = etree.SubElement(transition_list_tag, "transition",
                                                      {'id': transition.id,
                                                       'name': transition.name,
                                                       'type': str(transition.type),
                                                       'alias': transition.alias,
                                                       'source': str(transition.source),
                                                       'destination': str(transition.destination)})
                    _transition_part_list_tag = etree.SubElement(transition_tag, "conditionList")
        self.write()

    # Method to write transition's type by list [transition, type]
    def write_transition_type(self, transition_type_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition in transition_type_list:
                for transition_tag in root.findall(".//transition[@id='" + transition[0].id + "']"):
                    transition_tag.set('type', transition[1])
        self.write()

    # Method to write transition's alias by list [transition, alias]
    def write_transition_alias(self, transition_alias_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition in transition_alias_list:
                for transition_tag in root.findall(".//transition[@id='" + transition[0].id + "']"):
                    transition_tag.set('alias', transition[1])
        self.write()

    # Method to write transition's condition by list [transition, condition]
    def write_transition_condition(self, transition_condition_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition in root.findall(".//transition"):
                for tra, condition in transition_condition_list:
                    if transition.get('id') == tra.id:
                        tag = transition.find('conditionList')
                        state_part_tag = etree.SubElement(tag, "condition",
                                                          {'text': str(condition)})
        self.write()

    # Method to write transition's source by list [transition, source]
    def write_source(self, transition_source_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition_src in transition_source_list:
                for state_tag in root.findall(".//transition[@id='" + transition_src[0].id + "']"):
                    state_tag.set('source', transition_src[1].id)
        self.write()

    # Method to write transition's destination by list [transition, destination]
    def write_destination(self, transition_destination_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition_dest in transition_destination_list:
                for state_tag in root.findall(".//transition[@id='" + transition_dest[0].id + "']"):
                    state_tag.set('destination', transition_dest[1].id)
        self.write()

    # Method to delete transition by list [transition]
    def delete_transition(self, delete_transition_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for transition in delete_transition_list:
                for transition_tag in root.findall(".//transition[@id='" + transition.id + "']"):
                    transition_tag.getparent().remove(transition_tag)
        self.write()

    # Method to write functional element from functional elements list
    def write_functional_element(self, functional_element_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('functionalElementList') is None:
                etree.SubElement(root, 'functionalElementList')
            for functional_element_list_tag in root.findall("./functionalElementList"):
                for functional_element in functional_element_list:
                    functional_element_tag = etree.SubElement(functional_element_list_tag,
                                                              "functionalElement", {
                                                                  'id': functional_element.id,
                                                                  'name': functional_element.name,
                                                                  'type': str(
                                                                      functional_element.type),
                                                                  'alias': functional_element.alias}
                                                              )

                    _fun_elem_part_list_tag = etree.SubElement(functional_element_tag,
                                                               "functionalElementPartList")
                    _allocated_state_list_tag = etree.SubElement(functional_element_tag,
                                                                 "allocatedStateList")
                    _allocated_function_list_tag = etree.SubElement(functional_element_tag,
                                                                    "allocatedFunctionList")
                    _exposed_interface_list_tag = etree.SubElement(functional_element_tag,
                                                                   "exposedInterfaceList")
        self.write()

    # Method to write child by list [parent, child]
    def write_functional_element_child(self, fun_elem_child_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each flow/data
            for functional_element in root.findall(".//functionalElement"):
                for parent, child in fun_elem_child_list:
                    if functional_element.get('id') == parent.id:
                        tag = functional_element.find('functionalElementPartList')
                        _functional_element_part_tag = etree.SubElement(tag,
                                                                        "functionalElementPart",
                                                                        {'id': child.id})
        self.write()

    # Method to write allocated state by list [fun_elem, allocated_state]
    def write_allocated_state(self, fun_elem_state_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each flow/data
            for functional_element in root.findall(".//functionalElement"):
                for fun_elem, state in fun_elem_state_list:
                    if functional_element.get('id') == fun_elem.id:
                        tag = functional_element.find('allocatedStateList')
                        allocated_state_tag = etree.SubElement(tag, "allocatedState",
                                                               {'id': state.id})
        self.write()

    # Method to write allocated function by list [fun_elem, allocated_function]
    def write_allocated_function(self, fun_elem_function_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each flow/data
            for functional_element in root.findall(".//functionalElement"):
                for fun_elem, function in fun_elem_function_list:
                    if functional_element.get('id') == fun_elem.id:
                        tag = functional_element.find('allocatedFunctionList')
                        _allocated_state_tag = etree.SubElement(tag, "allocatedFunction",
                                                                {'id': function.id})
        self.write()

    # Method to write exposed interfaces by list [fun_elem, exposed_interface]
    def write_exposed_interface(self, fun_elem_inter_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for functional_element in root.findall(".//functionalElement"):
                if functional_element.find('exposedInterfaceList') is None:
                    etree.SubElement(functional_element, 'exposedInterfaceList')

                for fun_elem, inter in fun_elem_inter_list:
                    if functional_element.get('id') == fun_elem.id:
                        tag = functional_element.find('exposedInterfaceList')
                        _exposed_interface_tag = etree.SubElement(tag, "exposedInterface",
                                                                  {'id': inter.id})
        self.write()

    # Method to delete functional element by list [functional element]
    def delete_functional_element(self, delete_fun_elem_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for fun_elem in delete_fun_elem_list:
                for fun_elem_tag in root.findall(".//functionalElement[@id='" + fun_elem.id + "']"):
                    fun_elem_tag.getparent().remove(fun_elem_tag)
        self.write()

    # Method to write fun_elem's type by list [fun_elem, type]
    def write_fun_elem_type(self, fun_elem_type_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for fun_elem in fun_elem_type_list:
                for fun_elem_tag in root.findall(".//functionalElement[@id='" + fun_elem[0].id
                                                 + "']"):
                    fun_elem_tag.set('type', fun_elem[1])
        self.write()

    # Method to write fun_elem's alias by list [fun_elem, alias]
    def write_fun_elem_alias(self, fun_elem_alias_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for fun_elem in fun_elem_alias_list:
                for fun_elem_tag in root.findall(".//functionalElement[@id='" + fun_elem[0].id
                                                 + "']"):
                    fun_elem_tag.set('alias', fun_elem[1])
        self.write()

    # Method to write chains from chain's list
    def write_chain(self, chain_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('chainList') is None:
                etree.SubElement(root, 'chainList')
            for chain_list_tag in root.findall("./chainList"):
                for chain in chain_list:
                    chain_tag = etree.SubElement(chain_list_tag, "chain",
                                                 {'id': chain.id, 'name': chain.name,
                                                  'type': str(chain.type)})
                    _allocated_item_list_tag = etree.SubElement(chain_tag, "allocatedItemList")
        self.write()

    # Method to write allocated item by list [chain, allocated_item]
    def write_allocated_chain_item(self, chain_item_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each flow/data
            for chain_element in root.findall(".//chain"):
                for chain, item in chain_item_list:
                    if chain_element.get('id') == chain.id:
                        tag = chain_element.find('allocatedItemList')
                        _allocated_item_tag = etree.SubElement(tag, "allocatedItem",
                                                               {'id': item.id})
        self.write()

    # Method to write attributes from attribute's list
    def write_attribute(self, attribute_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('attributeList') is None:
                etree.SubElement(root, 'attributeList')
            for attribute_list_tag in root.findall("./attributeList"):
                for attribute in attribute_list:
                    attribute_tag = etree.SubElement(attribute_list_tag, "attribute",
                                                     {'id': attribute.id, 'name': attribute.name,
                                                      'type': str(attribute.type),
                                                      'alias': attribute.alias})

                    _described_item_list_tag = etree.SubElement(attribute_tag, "describedItemList")

        self.write()

    # Method to write described item by list [attribute, (described_item, value)]
    def write_described_attribute_item(self, attribute_item_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each attribute
            for attribute_element in root.findall(".//attribute"):
                for attribute, item in attribute_item_list:
                    if attribute_element.get('id') == attribute.id:
                        tag = attribute_element.find('describedItemList')
                        _allocated_item_tag = etree.SubElement(tag, "describedItem",
                                                               {'id': item[0].id, 'value': item[1]})
        self.write()

    # Method to write attribute's type by list [Attribute, type_str]
    def write_attribute_type(self, attribute_type_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for attribute in attribute_type_list:
                for attribute_tag in root.findall(".//attribute[@id='" + attribute[0].id
                                                  + "']"):
                    attribute_tag.set('type', attribute[1])
        self.write()

    # Method to write interfaces from interface's list
    def write_functional_interface(self, functional_interface_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            if root.find('functionalInterfaceList') is None:
                etree.SubElement(root, 'functionalInterfaceList')
            for fun_interface_list_tag in root.findall("./functionalInterfaceList"):
                for fun_interface in functional_interface_list:
                    fun_interface_tag = etree.SubElement(fun_interface_list_tag,
                                                         "functionalInterface",
                                                         {'id': fun_interface.id,
                                                          'name': fun_interface.name,
                                                          'type': str(fun_interface.type),
                                                          'alias': fun_interface.alias})
                    _allocated_data_list_tag = etree.SubElement(fun_interface_tag,
                                                                "allocatedDataList")
        self.write()

    # Method to write allocated data by list [Functional Interface, allocated_data]
    def write_fun_interface_allocated_data(self, fun_inter_data_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            # Loop on each flow/data
            for fun_interface_tag in root.findall(".//functionalInterface"):
                for fun_interface, data in fun_inter_data_list:
                    if fun_interface_tag.get('id') == fun_interface.id:
                        tag = fun_interface_tag.find('allocatedDataList')
                        _allocated_data_tag = etree.SubElement(tag, "allocatedData",
                                                               {'id': data.id})
        self.write()

    # Method to write functional interface's type by list [Fun interface, type_str]
    def write_fun_interface_type(self, fun_inter_type_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for fun_inter_type in fun_inter_type_list:
                for fun_inter_tag in root.findall(".//functionalInterface[@id='" +
                                                  fun_inter_type[0].id + "']"):
                    fun_inter_tag.set('type', fun_inter_type[1])
        self.write()

    # Method to write fun_inter's alias by list [fun_inter, alias]
    def write_fun_interface_alias(self, fun_inter_alias_list):
        with open(self.file, 'rb') as file:
            parser = etree.XMLParser(remove_blank_text=True)
            root = self.tree.parse(file, parser)
            for fun_inter_alias in fun_inter_alias_list:
                for fun_inter_tag in root.findall(".//functionalInterface[@id='" +
                                                  fun_inter_alias[0].id + "']"):
                    fun_inter_tag.set('alias', fun_inter_alias[1])
        self.write()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Libraries
import re
import os
import shutil
import getpass
import pandas as pd
from datetime import datetime
from io import StringIO

from IPython.core.magic import (Magics, magics_class, cell_magic)
from IPython.display import display, HTML, Markdown

# Modules
from . import orchestrator
from . import question_answer
import xml_adapter


# The class MUST call this class decorator at creation time
@magics_class
class MyMagics(Magics):

    @cell_magic
    def jarvis(self, line, cell):
        # We create a string buffer containing the
        # contents of the cell.
        sio = StringIO(cell)
        # Take the value within the buffer
        input_str = sio.getvalue()
        # Delete the '"' from input string, to avoid xml to plantuml errors.
        input_str = input_str.replace('"', "")
        # Delete extra whitespaces
        input_str = re.sub(' +', ' ', input_str)
        # Get model's declaration, need a space after "with" otherwise print a message
        xml_name_str = re.match(r"^with (.*)(?=.|\n)", input_str, re.MULTILINE)
        if xml_name_str:
            xml_name = xml_name_str.group(1)
            # If the model(i.e. file) already exists, parse it to extract lists
            if os.path.isfile(f"{xml_name}.xml"):
                xml_lists = xml_adapter.parse_xml(f"{xml_name}.xml")
                print(f"{xml_name}.xml parsed")
                xml_function_list = xml_lists[0]
                xml_consumer_function_list = xml_lists[1]
                xml_producer_function_list = xml_lists[2]
                xml_function_parent_dict = xml_lists[3]
                xml_data_list = xml_lists[4]
                xml_state_list = xml_lists[5]
                xml_state_parent_dict = xml_lists[6]
                xml_transition_list = xml_lists[7]
                xml_fun_elem_list = xml_lists[8]
                xml_fun_elem_parent_dict = xml_lists[9]
                xml_chain_list = xml_lists[10]
                xml_attribute_list = xml_lists[11]
                xml_fun_inter_list = xml_lists[12]
                output_xml = xml_adapter.generate_xml(f"{xml_name}.xml")
            # Else create an empty xml named by "xml_name"(and associated empty lists)
            # or will be named by default "Outpout"
            else:
                xml_function_list = set()
                xml_consumer_function_list = []
                xml_producer_function_list = []
                xml_function_parent_dict = {}
                xml_data_list = set()
                xml_state_list = set()
                xml_state_parent_dict = {}
                xml_transition_list = set()
                xml_fun_elem_list = set()
                xml_fun_elem_parent_dict = {}
                xml_chain_list = set()
                xml_attribute_list = set()
                xml_fun_inter_list = set()
                if len(xml_name) > 1:
                    print(f"Creating {xml_name}.xml !")
                    output_xml = xml_adapter.generate_xml(f"{xml_name}.xml")
                    output_xml.write()
                else:
                    print("Xml's file does not exists, creating it('output.xml' by default) !")
                    output_xml = xml_adapter.generate_xml("")
                    output_xml.write()

            xml_dict = {'xml_function_list': xml_function_list,
                        'xml_consumer_function_list': xml_consumer_function_list,
                        'xml_producer_function_list': xml_producer_function_list,
                        'xml_function_parent_dict': xml_function_parent_dict,
                        'xml_data_list': xml_data_list,
                        'xml_state_list': xml_state_list,
                        'xml_state_parent_dict': xml_state_parent_dict,
                        'xml_transition_list': xml_transition_list,
                        'xml_fun_elem_list': xml_fun_elem_list,
                        'xml_fun_elem_parent_dict': xml_fun_elem_parent_dict,
                        'xml_chain_list': xml_chain_list,
                        'xml_attribute_list': xml_attribute_list,
                        'xml_fun_inter_list': xml_fun_inter_list,
                        'output_xml': output_xml,
                        'xml_name': xml_name}

            update = lookup(input_str, LOOKUPS, **xml_dict)

            if not update:
                return
            else:
                if 1 in update:
                    self.show_model_update_msg(xml_name)
                else:
                    self.show_no_model_update_msg(xml_name)
        else:
            print(
                "Bad model's declaration, model's name should be written or add a ' '(blank space) "
                "after 'with' command to create default 'Output.xml'")

    @classmethod
    def show_model_update_msg(cls, xml_name):
        print(f"{xml_name}.xml updated")

    @classmethod
    def show_no_model_update_msg(cls, xml_name):
        print(f"No update for {xml_name}.xml")


LOOKUPS = [
    (r"(under.*)",
     lambda matched_str, **kwargs: matched_under(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is a function\b(?=.|\n)",
     lambda matched_str, **kwargs: matched_function(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is a data(?=.\s|\n|.\n)",
     lambda matched_str, **kwargs: matched_data(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is a state(?=.|\n)",
     lambda matched_str, **kwargs: matched_state(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is a transition(?=.|\n)",
     lambda matched_str, **kwargs: matched_transition(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is a functional element(?=.|\n)",
     lambda matched_str, **kwargs: matched_functional_element(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is a functional interface(?=.|\n)",
     lambda matched_str, **kwargs: matched_functional_interface(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is an attribute\b(?=.|\n)",
     lambda matched_str, **kwargs: matched_attribute(matched_str, **kwargs)),

    (r"The alias of (.*?) is ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_alias(matched_str, **kwargs)),

    (r"(?<= |\n)consider ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_consider(matched_str, **kwargs)),

    (r"([^\. \.\n]*) is composed of ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_composition(matched_str, **kwargs)),

    (r"([^\. \.\n]*) composes ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_composition(reverse(matched_str), **kwargs)),

    (r"([^\. \.\n]*) consumes ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_consumer(reverse(matched_str), **kwargs)),

    (r"([^\. \.\n]*) is an input of ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_consumer(matched_str, **kwargs)),

    (r"([^\. \.\n]*) produces ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_producer(reverse(matched_str), **kwargs)),

    (r"([^\. \.\n]*) is an output of ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_producer(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) exposes ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_exposes(matched_str, **kwargs)),

    (r"(?<= |\n)(.*?) is allocated to ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_allocation(reverse(matched_str), **kwargs)),

    (r"(?<= |\n)(.*?) allocates ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_allocation(matched_str, **kwargs)),

    (r"(?<= |\n)delete ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_delete(matched_str, **kwargs)),

    (r"The type of (.*?) is ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_type(matched_str, **kwargs)),

    (r"([^\. \.\n]*) implies ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_implies(matched_str, **kwargs)),

    (r"Condition for (.*?) is:([^\.\n]*)",
     lambda matched_str, **kwargs: matched_condition(matched_str, **kwargs)),

    (r"The (source|destination) of (.*?) is ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_src_dest(matched_str, **kwargs)),

    (r"(?<= |\n)show (.*?)\n",
     lambda matched_str, **kwargs: matched_show(matched_str, **kwargs)),

    (r"\s([A-Za-z\s].*\?)",
     lambda matched_str, **kwargs: matched_question_mark(matched_str, **kwargs)),

    (r"list (input|output|child|data|function|transition|interface) ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_list(matched_str, **kwargs)),

    (r"The ([^type|alias|source|destination].*?) of (.*?) is ([^\.\n]*)",
     lambda matched_str, **kwargs: matched_described_attribute(matched_str, **kwargs)),
]


def lookup(string, lookups, **kwargs):
    """Lookup table with conditions depending on the match"""
    update_list = []
    for regex, values in lookups:
        result_chain = None
        result = None
        update = None
        if regex == r"(under.*)":
            result_chain = re.split(regex, string)
            del result_chain[0]
        # Only one diagram per cell can be output
        elif regex == r"(?<= |\n)show (.*?)\n":
            result = re.search(regex, string, re.MULTILINE)
        else:
            # Transform to avoid duplicated function's declaration within cells input
            result = []
            [result.append(x) for x in re.findall(regex, string, re.MULTILINE) if x not in result]

        if result and not result_chain:
            update = values(result, **kwargs)
        elif result_chain:
            string = ''
            update = matched_under(result_chain, **kwargs)

        if update and isinstance(update, list):
            update_list.append(*update)

    return update_list


def matched_under(chain_name_str, **kwargs):
    """Get "under" declaration"""
    out = []
    for a, b in zip(chain_name_str[::2], chain_name_str[1::2]):
        a = a.replace("under ", "")
        out += orchestrator.add_chain(a,
                                      kwargs['xml_chain_list'],
                                      kwargs['output_xml'])
        lookup(b, LOOKUPS, **kwargs)
    if 1 in out:
        return [1]
    else:
        return [0]


def matched_function(function_name_str_list, **kwargs):
    """Get function's declaration (does not match if "function" is not at the end)"""
    out = orchestrator.add_function_by_name(function_name_str_list,
                                            kwargs['xml_function_list'],
                                            kwargs['output_xml'])
    return out


def matched_data(data_str_list, **kwargs):
    """Get data declaration"""
    out = orchestrator.add_data(data_str_list, kwargs['xml_data_list'],
                                kwargs['output_xml'])
    return out


def matched_state(state_name_str_list, **kwargs):
    """Get state's declaration"""
    out = orchestrator.add_state_by_name(state_name_str_list,
                                         kwargs['xml_state_list'],
                                         kwargs['output_xml'])
    return out


def matched_transition(transition_name_str_list, **kwargs):
    """Get transition's declaration"""
    out = orchestrator.add_transition_by_name(transition_name_str_list,
                                              kwargs['xml_transition_list'],
                                              kwargs['output_xml'])
    return out


def matched_functional_element(functional_elem_name_str_list, **kwargs):
    """Get Functional element's declaration"""
    out = orchestrator.add_fun_elem_by_name(functional_elem_name_str_list,
                                            kwargs['xml_fun_elem_list'],
                                            kwargs['output_xml'])
    return out


def matched_functional_interface(functional_inter_name_str_list, **kwargs):
    """Get Functional interface's declaration"""
    out = orchestrator.add_fun_inter_by_name(functional_inter_name_str_list,
                                             kwargs['xml_fun_inter_list'],
                                             kwargs['output_xml'])
    return out


def matched_attribute(attribute_name_str, **kwargs):
    """Get "attribute" declaration"""
    out = orchestrator.add_attribute(attribute_name_str,
                                     kwargs['xml_attribute_list'],
                                     kwargs['output_xml'])
    return out


def matched_alias(alias_str_list, **kwargs):
    """Get "alias" declaration"""
    out = orchestrator.check_set_object_alias(alias_str_list,
                                              kwargs['xml_function_list'],
                                              kwargs['xml_state_list'],
                                              kwargs['xml_transition_list'],
                                              kwargs['xml_fun_elem_list'],
                                              kwargs['xml_fun_inter_list'],
                                              kwargs['output_xml'])
    return out


def matched_consider(consider_str_list, **kwargs):
    """Get "consider" declaration"""
    out = orchestrator.check_get_consider(consider_str_list,
                                          kwargs['xml_function_list'],
                                          kwargs['xml_fun_elem_list'],
                                          kwargs['xml_data_list'],
                                          kwargs['xml_chain_list'],
                                          kwargs['output_xml'])
    return out


def matched_composition(parent_child_name_str_list, **kwargs):
    """Get composition relationship command (match for "is composed by' or "composes")"""
    out = orchestrator.check_add_child(parent_child_name_str_list,
                                       kwargs['xml_function_list'],
                                       kwargs['xml_function_parent_dict'],
                                       kwargs['xml_state_list'],
                                       kwargs['xml_state_parent_dict'],
                                       kwargs['xml_fun_elem_list'],
                                       kwargs['xml_fun_elem_parent_dict'],
                                       kwargs['output_xml'])
    return out


def matched_consumer(consumer_str_list, **kwargs):
    """Get consumer declaration"""
    out = orchestrator.check_add_consumer_function(consumer_str_list,
                                                   kwargs['xml_consumer_function_list'],
                                                   kwargs['xml_producer_function_list'],
                                                   kwargs['xml_function_list'],
                                                   kwargs['xml_data_list'],
                                                   kwargs['output_xml'])
    return out


def matched_producer(producer_str_list, **kwargs):
    """Get producer declaration"""
    out = orchestrator.check_add_producer_function(producer_str_list,
                                                   kwargs['xml_consumer_function_list'],
                                                   kwargs['xml_producer_function_list'],
                                                   kwargs['xml_function_list'],
                                                   kwargs['xml_data_list'],
                                                   kwargs['output_xml'])
    return out


def matched_allocation(allocation_str_list, **kwargs):
    """Get allocation declaration"""
    out = orchestrator.check_add_allocation(allocation_str_list,
                                            kwargs['xml_fun_elem_list'],
                                            kwargs['xml_state_list'],
                                            kwargs['xml_function_list'],
                                            kwargs['xml_fun_inter_list'],
                                            kwargs['xml_data_list'],
                                            kwargs['xml_consumer_function_list'],
                                            kwargs['xml_producer_function_list'],
                                            kwargs['output_xml'])
    return out


def matched_exposes(exposes_str_list, **kwargs):
    """Get 'exposes' declaration"""
    out = orchestrator.check_add_exposes(exposes_str_list,
                                         kwargs['xml_fun_elem_list'],
                                         kwargs['xml_fun_inter_list'],
                                         kwargs['xml_data_list'],
                                         kwargs['output_xml'])
    return out


def matched_delete(delete_str_list, **kwargs):
    """Get delete declaration"""
    out = orchestrator.check_and_delete(delete_str_list,
                                        kwargs['xml_function_list'],
                                        kwargs['xml_producer_function_list'],
                                        kwargs['xml_consumer_function_list'],
                                        kwargs['xml_data_list'],
                                        kwargs['xml_state_list'],
                                        kwargs['xml_transition_list'],
                                        kwargs['xml_fun_elem_list'],
                                        kwargs['output_xml'])
    return out


def matched_type(type_str_list, **kwargs):
    """Get set_type declaration"""
    out = orchestrator.check_set_object_type(type_str_list,
                                             kwargs['xml_function_list'],
                                             kwargs['xml_data_list'],
                                             kwargs['xml_state_list'],
                                             kwargs['xml_transition_list'],
                                             kwargs['xml_fun_elem_list'],
                                             kwargs['xml_attribute_list'],
                                             kwargs['xml_fun_inter_list'],
                                             kwargs['output_xml'])
    return out


def matched_implies(data_predecessor_str_set, **kwargs):
    """Get predecessor declaration"""
    out = orchestrator.check_add_predecessor(data_predecessor_str_set,
                                             kwargs['xml_data_list'],
                                             kwargs['xml_chain_list'],
                                             kwargs['output_xml'])
    return out


def matched_condition(condition_str_list, **kwargs):
    """Get set_condition declaration"""
    out = orchestrator.check_add_transition_condition(condition_str_list,
                                                      kwargs['xml_transition_list'],
                                                      kwargs['output_xml'])
    return out


def matched_src_dest(src_dest_str, **kwargs):
    """Get source/destination declaration for transition"""
    out = orchestrator.check_add_src_dest(src_dest_str,
                                          kwargs['xml_transition_list'],
                                          kwargs['xml_state_list'],
                                          kwargs['output_xml'])
    return out


# Get "show" declaration
def matched_show(diagram_name_str, **kwargs):
    """Get "show" declaration"""
    out = orchestrator.filter_show_command(diagram_name_str, **kwargs)
    if out:
        hyper = get_hyperlink(out)
        display(HTML(hyper))
        print("Overview :")
        return display(Markdown(f'![figure]({out})'))


def matched_question_mark(question_str, **kwargs):
    """Gets "?" declaration"""
    out = question_answer.find_question(question_str, **kwargs)
    if out:
        for elem in out:
            if isinstance(elem, str):
                print(elem)
            else:
                display(elem)


def matched_list(object_str, **kwargs):
    """Gets list declaration"""
    out = question_answer.get_object_list(object_str, **kwargs)
    if out:
        for i in out:
            if "Child" in i[0] or "Function" in i[0]:
                title = i.pop(0)
                df = pd.DataFrame(i, columns=["Object's name", "Relationship's type"])
                df = df.T
                # Could be usefull to add it with button next to table but needs ipywidgets ...
                # df.to_clipboard(excel=True)
                df = df.style\
                    .set_caption(title)\
                    .set_properties(**{'white-space': 'nowrap'})
                display(df)
            elif "Input" in i[0]:
                title = i.pop(0)
                df = pd.DataFrame(i, columns=["Data name", "Producer"])
                df = df.T
                df = df.style\
                    .set_caption(title)\
                    .set_properties(**{'white-space': 'nowrap'})
                df = df.to_html().replace("\\n", "<br>")
                display(HTML(df))
            elif "Output" in i[0]:
                title = i.pop(0)
                df = pd.DataFrame(i, columns=["Data name", "Consumer"])
                df = df.T
                df = df.style\
                    .set_caption(title)\
                    .set_properties(**{'white-space': 'nowrap'})
                df = df.to_html().replace("\\n", "<br>")
                display(HTML(df))
            elif "Data" in i[0] or "Transition" in i[0]:
                if "Transition" in i[0]:
                    first = 3
                    last = 4
                else:
                    first = 1
                    last = 5
                title = i.pop(0)
                df = pd.DataFrame(i)
                df = df.T
                for idx in range(first, last):
                    df.iloc[idx] = df.iloc[idx].str.join("\\n")
                df = df.style\
                    .set_caption(title)\
                    .set_properties(**{'white-space': 'nowrap'})
                df = df.to_html().replace("\\n", "<br>")
                display(HTML(df))
            elif "Interface" in i[0]:
                title = i.pop(0)
                df = pd.DataFrame(i, columns=["Interface ", "Last connected functional element"])
                df = df.T
                df = df.style\
                    .set_caption(title)\
                    .set_properties(**{'white-space': 'nowrap'})
                df = df.to_html().replace("\\n", "<br>")
                display(HTML(df))


def matched_described_attribute(described_attribute_str, **kwargs):
    """Get the described attribute value for an object"""
    out = orchestrator.check_add_object_attribute(described_attribute_str,
                                                  kwargs['xml_attribute_list'],
                                                  kwargs['xml_function_list'],
                                                  kwargs['xml_fun_elem_list'],
                                                  kwargs['xml_fun_inter_list'],
                                                  kwargs['output_xml'])
    return out


def reverse(inverted_list):
    """Reverses input tuple strings"""
    sorted_list = []
    for tuples in inverted_list:
        new_tup = ()
        for k in reversed(tuples):
            new_tup = new_tup + (k,)
        sorted_list.append(new_tup)
    return sorted_list


def greet_user():
    """Greets the user according to the time thanks to :
    https://ireadblog.com/posts/141/how-to-build-your-personal-ai-assistant-using-python"""
    hour = datetime.now().hour
    # Use getpass() because available on Unix/Windows
    user_name = getpass.getuser()
    if 6 <= hour < 12:
        print(f"Good Morning {user_name}")
    elif 12 <= hour < 16:
        print(f"Good afternoon {user_name}")
    elif 16 <= hour < 19:
        print(f"Good Evening {user_name}")
    print("I am Jarvis. How may I assist you?")


def clean_diagram_folder():
    """Clean/erase all files within Diagram's folder"""
    folder = './Diagrams'
    if not os.path.exists(folder):
        os.mkdir(folder)
    else:
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.remove(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                print('Failed to delete %s. Reason: %s' % (file_path, e))


def get_hyperlink(path):
    """Convert file path into clickable form."""
    text = "Click to open in new tab"
    # convert the url into link
    return '<a href="{}" target="_blank">{}</a>'.format(path, text)
